# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import scipy.io
import numpy as np
from scipy import ndimage
import weles
import operator

"""
A class to store and to use hypersectral image.
"""
class HS:
    """
    ## Initialization
    """
    def __init__(self, dictionary):
        # Loading image, ground truth and establishing informations.
        self.image = self.loadMatFromTuple(dictionary['image'])
        self.gt = self.loadMatFromTuple(dictionary['gt'])
        self.name = dictionary['name']
        self.classes = dictionary['classes']

        #print type(self.image[(0,0)][0])

        # Getting in shape.
        shape = np.shape(self.image)
        self.rows = shape[0]
        self.cols = shape[1]
        self.bands = shape[2]

        # Normalize
        self.normA = []
        self.normB = []
        for sliceIdx in xrange(self.bands):
            slice = self.slice(sliceIdx)
            #print sliceIdx
            minimum = np.min(slice)
            maximum = np.max(slice)
            self.normA.append(minimum)
            self.normB.append(maximum)
            #print slice
            #print 'min %i max %i' % (minimum, maximum)
        self.normB = map(operator.sub, self.normB, self.normA)
        #print self.normA
        #print self.normB


        # Searching for maximum value
        self.max = np.amax(self.image)
        self.maxlabel = np.amax(self.gt)

        self.reverseClasses = None
        self.prepareReverse()

    ## Operators

    def signatures(self):
        labels = set()
        result = {}
        # Establish list of labels, according to GT
        for row in self.gt:
            for item in row:
                if item not in labels:
                    labels.add(item)
        # Calculate mean signature for every label
        for label in labels:
            signatures = []
            for x, row in enumerate(self.gt):
                for y, item in enumerate(row):
                    if item == label:
                        signatures.append(self.signature((x, y)))
            result[label] = np.mean(signatures, axis = 0)
        # return the result
        return result

    def signaturesPNG(self, filename):
        signatures = self.signatures()

        # use LaTeX
        #plt.rc('text', usetex=True)
        #plt.rc('font', family='serif')
        #rcParams['font.sans-serif'] = ['Tahoma']

        # Plot size
        plt.figure(figsize=(11, 5.5))

        # Remove the plot frame lines. They are unnecessary chartjunk.
        ax = plt.subplot(111)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)

        # Limit the range of the plot to only where the data is.
        # Avoid unnecessary whitespace.
        plt.ylim(0, 1)
        plt.xlim(0, self.bands)

        # Make sure your axis ticks are large enough to be easily read.
        # You don't want your viewers squinting to read your plot.
        plt.yticks(np.arange(0, 1, .1), [str(x) + "%" for x in range(0, 91, 10)], fontsize=14)
        plt.xticks(fontsize=14)

        # Provide tick lines across the plot to help your viewers trace along
        # the axis ticks. Make sure that the lines are light and small so they
        # don't obscure the primary data lines.
        for y in np.arange(.1, .91, .1):
            plt.plot(range(self.bands), [y] * len(range(self.bands)), "--", lw=0.5, color="black", alpha=0.3)

        # Remove the tick marks; they are unnecessary with the tick lines we just plotted.
        plt.tick_params(axis="both", which="both", bottom="on", top="off",
                        labelbottom="on", left="off", right="off", labelleft="on")

        # Actually plot signatures
        for item in signatures:
            line, = plt.plot(
                xrange(self.bands),
                signatures[item],
                linewidth = 1 )
            line.set_antialiased(True)

        plt.xlabel('Band')
        plt.ylabel('Normalized reflectance')

        plt.savefig(filename, bbox_inches='tight')

    """
    ### Getting sample
    """
    def sample(self, location):
        return weles.Sample(
            self.signature(location),
            self.label(location))

    """
    ### Getting signature
    """
    def signature(self, location):
        return (self.image[location].astype(float) - self.normA) / self.normB

    """
    ### Getting label
    """
    def label(self, location):
        return int(np.copy(self.gt[location]))

    """
    ### Getting slice
    """
    def slice(self, band):
        return np.copy(self.image[:, :, band])

    ## Helper functions

    """
    ## Entropodynamic Percentyle Filter
    """
    def dynamicsTensor(self, slice, ksize):
        layers = ksize[0] * ksize[1]
        tensor = np.zeros((self.rows, self.cols, layers))
        i = 0

        # Moving Spermatozoid
        for x in xrange(ksize[0]):
            for y in xrange(ksize[1]):
                kernel = np.zeros(ksize)
                kernel[x,y] = 1.

                smoothed = ndimage.convolve(slice, kernel, mode='reflect', cval=0.0)
                tensor[:,:,i] = smoothed
                i += 1

        return tensor

    def edges2(self, map, ksize):
        # Establish tensor
        tensor = self.dynamicsTensor(map, ksize)

        # Calculate denivelation
        edges2 = np.subtract(
            np.amax(tensor, axis=2),
            np.amin(tensor, axis=2)
        )

        # Normalize
        a = np.min(edges2)
        b = np.max(edges2)
        edges2 = np.divide(np.subtract(edges2, a), b - a)

        return edges2

    def edges3(self, ksize):
        edges3 = np.zeros(np.shape(self.image))
        for sid in xrange(self.bands):
            # Gather slice
            slice = self.slice(sid)

            # Assign
            edges3[:,:,sid] = self.edges2(slice, ksize)

        return edges3

    def epf(self, edges3d, percentile):
        # Calculate entropy
        entropy = np.percentile(
            np.percentile(
                edges3d, percentile, axis = 0),
            percentile, axis = 0)
        entropy = np.absolute(
            np.subtract(
                entropy, np.median(entropy)))

        # mean entropy filter
        mef = entropy < np.percentile(entropy, percentile)
        meanEntropy = np.zeros(len(entropy))
        meanEntropy[mef] = True

        # Entropy dynamics
        entropyDynamics = np.copy(entropy)
        val = entropyDynamics[0]
        entropyDynamics = np.delete(entropyDynamics, 0)
        entropyDynamics = np.append(entropyDynamics, entropyDynamics[-1])
        entropyDynamics = np.subtract(entropy, entropyDynamics)
        entropyDynamics = np.absolute(entropyDynamics)

        # Mean dynamics filter
        med = entropyDynamics < np.percentile(entropyDynamics, percentile)
        meanDynamics = np.zeros(len(entropy))
        meanDynamics[med] = True

        # Union filter
        union = [a and b for a,b in zip(meanEntropy, meanDynamics)]

        return (entropy, meanEntropy, entropyDynamics, meanDynamics, union)

    def bordersMap(self, source, filter):
        filteredEdges = np.squeeze(source[:,:,np.where(filter)])

        bordersMap = np.max(filteredEdges, 2)
        bordersMap = scipy.ndimage.median_filter(
            bordersMap,
            size=(2, 2)
        )
        return bordersMap

    def bordersMask(self, edgesFlat, percentile = 75):
        edgesMask = np.zeros(np.shape(edgesFlat))
        lvi = edgesFlat > np.percentile(edgesFlat, percentile)
        edgesMask[lvi] = True
        edgesMask = ndimage.binary_dilation(edgesMask)
        return edgesMask.astype(bool)

    """
    ## Loading from .mat file
    """
    def loadMatFromTuple(self, entry):
        return scipy.io.loadmat(entry[0])[entry[1]]

    def prepareReverse(self):
        self.reverseClasses = {}
        for (x,y), value in np.ndenumerate(self.gt):
            label = self.label((x, y))
            if not label in self.reverseClasses:
                self.reverseClasses.update({label: len(self.reverseClasses)})

    """
    ## Exporting Weles dataset
    """
    def dataset(self):
        dataset = weles.Dataset()
        db_name = self.name
        source_samples = []
        self.reverseClasses = {}
        for x in xrange(self.rows):
            for y in xrange(self.cols):
                sample = self.sample((x, y))
                source_samples.append(sample)
                if not sample.label in self.reverseClasses:
                    self.reverseClasses.update({sample.label: len(self.reverseClasses)})
                sample.label = self.reverseClasses[sample.label]
        dataset.fill(
            db_name,
            source_samples,
            self.reverseClasses
        )
        print '# Classes\n%s' % self.reverseClasses
        return dataset

    """
    ## Be verbose, man
    """
    def __str__(self):
        return '%s image, %i classes, %i samples of %i bands' % (
            self.name,
            len(self.classes),
            self.rows * self.cols,
            self.bands)



# -*- coding: utf-8 -*-

class AP:
    def __init__(self, img, ksize, percentile):
        self.img = img
        self.ksize = ksize
        self.percentile = percentile

        # Kostka graniczna
        self.edges3 = img.edges3(self.ksize)

        # Filtr
        epf = img.epf(self.edges3, percentile)

        self.entropy = epf[0]
        self.meanEntropy = epf[1]
        self.entropyDynamics = epf[2]
        self.meanDynamics = epf[3]

        self.filter = epf[4]

    def bordersMap(self, filter = None):
        if not filter:
            filter = self.filter
        bordersMap = self.img.bordersMap(self.edges3, filter)
        return bordersMap

    def weird(self, filter = None):
        if not filter:
            filter = self.filter
        weird = self.img.edges2(self.bordersMap(filter), self.ksize)
        return weird

    def bordersMask(self, filter = None, percentile = 75):
        if not filter:
            filter = self.filter
        bordersMask = self.img.bordersMask(self.bordersMap(filter), percentile)
        return bordersMask



import numpy as np
import warnings

def anisodiff(img,niter=1,kappa=50,gamma=0.1,step=(1.,1.),option=1,ploton=False):
	"""
	Anisotropic diffusion.

	Usage:
	imgout = anisodiff(im, niter, kappa, gamma, option)

	Arguments:
	        img    - input image
	        niter  - number of iterations
	        kappa  - conduction coefficient 20-100 ?
	        gamma  - max value of .25 for stability
	        step   - tuple, the distance between adjacent pixels in (y,x)
	        option - 1 Perona Malik diffusion equation No 1
	                 2 Perona Malik diffusion equation No 2
	        ploton - if True, the image will be plotted on every iteration

	Returns:
	        imgout   - diffused image.

	kappa controls conduction as a function of gradient.  If kappa is low
	small intensity gradients are able to block conduction and hence diffusion
	across step edges.  A large value reduces the influence of intensity
	gradients on conduction.

	gamma controls speed of diffusion (you usually want it at a maximum of
	0.25)

	step is used to scale the gradients in case the spacing between adjacent
	pixels differs in the x and y axes

	Diffusion equation 1 favours high contrast edges over low contrast ones.
	Diffusion equation 2 favours wide regions over smaller ones.

	Reference: 
	P. Perona and J. Malik. 
	Scale-space and edge detection using ansotropic diffusion.
	IEEE Transactions on Pattern Analysis and Machine Intelligence, 
	12(7):629-639, July 1990.

	Original MATLAB code by Peter Kovesi  
	School of Computer Science & Software Engineering
	The University of Western Australia
	pk @ csse uwa edu au
	<http://www.csse.uwa.edu.au>

	Translated to Python and optimised by Alistair Muldal
	Department of Pharmacology
	University of Oxford
	<alistair.muldal@pharm.ox.ac.uk>

	June 2000  original version.       
	March 2002 corrected diffusion eqn No 2.
	July 2012 translated to Python
	"""

	# ...you could always diffuse each color channel independently if you
	# really want
	if img.ndim == 3:
		warnings.warn("Only grayscale images allowed, converting to 2D matrix")
		img = img.mean(2)

	# initialize output array
	img = img.astype('float32')
	imgout = img.copy()

	# initialize some internal variables
	deltaS = np.zeros_like(imgout)
	deltaE = deltaS.copy()
	NS = deltaS.copy()
	EW = deltaS.copy()
	gS = np.ones_like(imgout)
	gE = gS.copy()

	# create the plot figure, if requested
	if ploton:
		import pylab as pl
		from time import sleep

		fig = pl.figure(figsize=(20,5.5),num="Anisotropic diffusion")
		ax1,ax2 = fig.add_subplot(1,2,1),fig.add_subplot(1,2,2)

		ax1.imshow(img,interpolation='nearest')
		ih = ax2.imshow(imgout,interpolation='nearest',animated=True)
		ax1.set_title("Original image")
		ax2.set_title("Iteration 0")

		fig.canvas.draw()

	for ii in xrange(niter):

		# calculate the diffs
		deltaS[:-1,: ] = np.diff(imgout,axis=0)
		deltaE[: ,:-1] = np.diff(imgout,axis=1)

		# conduction gradients (only need to compute one per dim!)
		if option == 1:
			gS = np.exp(-(deltaS/kappa)**2.)/step[0]
			gE = np.exp(-(deltaE/kappa)**2.)/step[1]
		elif option == 2:
			gS = 1./(1.+(deltaS/kappa)**2.)/step[0]
			gE = 1./(1.+(deltaE/kappa)**2.)/step[1]

		# update matrices
		E = gE*deltaE
		S = gS*deltaS

		# subtract a copy that has been shifted 'North/West' by one
		# pixel. don't as questions. just do it. trust me.
		NS[:] = S
		EW[:] = E
		NS[1:,:] -= S[:-1,:]
		EW[:,1:] -= E[:,:-1]

		# update the image
		imgout += gamma*(NS+EW)

		if ploton:
			iterstring = "Iteration %i" %(ii+1)
			ih.set_data(imgout)
			ax2.set_title(iterstring)
			fig.canvas.draw()
			# sleep(0.01)

	return imgout

def anisodiff3(stack,niter=1,kappa=50,gamma=0.1,step=(1.,1.,1.),option=1,ploton=False):
	"""
	3D Anisotropic diffusion.

	Usage:
	stackout = anisodiff(stack, niter, kappa, gamma, option)

	Arguments:
	        stack  - input stack
	        niter  - number of iterations
	        kappa  - conduction coefficient 20-100 ?
	        gamma  - max value of .25 for stability
	        step   - tuple, the distance between adjacent pixels in (z,y,x)
	        option - 1 Perona Malik diffusion equation No 1
	                 2 Perona Malik diffusion equation No 2
	        ploton - if True, the middle z-plane will be plotted on every 
	        	 iteration

	Returns:
	        stackout   - diffused stack.

	kappa controls conduction as a function of gradient.  If kappa is low
	small intensity gradients are able to block conduction and hence diffusion
	across step edges.  A large value reduces the influence of intensity
	gradients on conduction.

	gamma controls speed of diffusion (you usually want it at a maximum of
	0.25)

	step is used to scale the gradients in case the spacing between adjacent
	pixels differs in the x,y and/or z axes

	Diffusion equation 1 favours high contrast edges over low contrast ones.
	Diffusion equation 2 favours wide regions over smaller ones.

	Reference: 
	P. Perona and J. Malik. 
	Scale-space and edge detection using ansotropic diffusion.
	IEEE Transactions on Pattern Analysis and Machine Intelligence, 
	12(7):629-639, July 1990.

	Original MATLAB code by Peter Kovesi  
	School of Computer Science & Software Engineering
	The University of Western Australia
	pk @ csse uwa edu au
	<http://www.csse.uwa.edu.au>

	Translated to Python and optimised by Alistair Muldal
	Department of Pharmacology
	University of Oxford
	<alistair.muldal@pharm.ox.ac.uk>

	June 2000  original version.       
	March 2002 corrected diffusion eqn No 2.
	July 2012 translated to Python
	"""

	# ...you could always diffuse each color channel independently if you
	# really want
	if stack.ndim == 4:
		warnings.warn("Only grayscale stacks allowed, converting to 3D matrix")
		stack = stack.mean(3)

	# initialize output array
	stack = stack.astype('float32')
	stackout = stack.copy()

	# initialize some internal variables
	deltaS = np.zeros_like(stackout)
	deltaE = deltaS.copy()
	deltaD = deltaS.copy()
	NS = deltaS.copy()
	EW = deltaS.copy()
	UD = deltaS.copy()
	gS = np.ones_like(stackout)
	gE = gS.copy()
	gD = gS.copy()

	# create the plot figure, if requested
	if ploton:
		import pylab as pl
		from time import sleep

		showplane = stack.shape[0]//2

		fig = pl.figure(figsize=(20,5.5),num="Anisotropic diffusion")
		ax1,ax2 = fig.add_subplot(1,2,1),fig.add_subplot(1,2,2)

		ax1.imshow(stack[showplane,...].squeeze(),interpolation='nearest')
		ih = ax2.imshow(stackout[showplane,...].squeeze(),interpolation='nearest',animated=True)
		ax1.set_title("Original stack (Z = %i)" %showplane)
		ax2.set_title("Iteration 0")

		fig.canvas.draw()

	for ii in xrange(niter):

		# calculate the diffs
		deltaD[:-1,: ,:  ] = np.diff(stackout,axis=0)
		deltaS[:  ,:-1,: ] = np.diff(stackout,axis=1)
		deltaE[:  ,: ,:-1] = np.diff(stackout,axis=2)

		# conduction gradients (only need to compute one per dim!)
		if option == 1:
			gD = np.exp(-(deltaD/kappa)**2.)/step[0]
			gS = np.exp(-(deltaS/kappa)**2.)/step[1]
			gE = np.exp(-(deltaE/kappa)**2.)/step[2]
		elif option == 2:
			gD = 1./(1.+(deltaD/kappa)**2.)/step[0]
			gS = 1./(1.+(deltaS/kappa)**2.)/step[1]
			gE = 1./(1.+(deltaE/kappa)**2.)/step[2]

		# update matrices
		D = gD*deltaD
		E = gE*deltaE
		S = gS*deltaS

		# subtract a copy that has been shifted 'Up/North/West' by one
		# pixel. don't as questions. just do it. trust me.
		UD[:] = D
		NS[:] = S
		EW[:] = E
		UD[1:,: ,: ] -= D[:-1,:  ,:  ]
		NS[: ,1:,: ] -= S[:  ,:-1,:  ]
		EW[: ,: ,1:] -= E[:  ,:  ,:-1]

		# update the image
		stackout += gamma*(UD+NS+EW)

		if ploton:
			iterstring = "Iteration %i" %(ii+1)
			ih.set_data(stackout[showplane,...].squeeze())
			ax2.set_title(iterstring)
			fig.canvas.draw()
			# sleep(0.01)

	return stackout